/**
 * 
 * $Id: PictureUtil.java,v 1.15 2009/09/14 05:12:22 aoki Exp $
 */
package util;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Transparency;
import java.awt.color.CMMException;
import java.awt.color.ColorSpace;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.ComponentColorModel;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Iterator;
import java.util.Locale;

import javax.imageio.IIOException;
import javax.imageio.IIOImage;
import javax.imageio.ImageIO;
import javax.imageio.ImageReader;
import javax.imageio.ImageWriteParam;
import javax.imageio.ImageWriter;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.metadata.IIOMetadataNode;
import javax.imageio.spi.ImageReaderSpi;
import javax.imageio.stream.ImageInputStream;
import javax.imageio.stream.ImageOutputStream;
import javax.imageio.stream.MemoryCacheImageOutputStream;

import org.w3c.dom.NodeList;

/**
 * 摜𑀍삷ׂ̃[eBeBNXłB
 * 
 * @author s.aoki
 */
public final class PictureUtil {

    /** JPEGkiݒ̃ftHgl */
    private static final float DEFAULT_JPG_COMP_QUALITY = 0.75f;
    
    /**
     * ̃NXŃT|[g摜^Cv̗񋓌^łB
     */
    public enum SupportedImageType {
        /** JPEG̉摜ł */
        JPG,
        /** PNG`̉摜ł */
        PNG,
        /** GIF`̉摜ł */
        GIF
    }
    
    static {
        // ImageIÕ[N
        // http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=4513817
        ImageIO.setUseCache(false);
    }
    
    // CX^X͐Ȃ
    private PictureUtil() {
        // do nothing.
    }

    /**
     * in̓eǂݎĉ摜̎ނ𔻕ʂAԒZgq
     * ɂĕԂ܂Bin摜`łȂꍇnullԂ܂B
     * 
     * @param in ΏۂInputStream
     * @return in̉摜ʂ\鏬̊gqBnullB
     * @throws IOException ͗Oꍇ
     */
    public static String getDefaultFileSuffix(InputStream in) throws IOException {
        String suffix = null;
        for (Iterator<ImageReader> itr = ImageIO.getImageReaders(
                ImageIO.createImageInputStream(in)); itr.hasNext(); ) {
            ImageReader reader = itr.next();
            ImageReaderSpi readerSpi = reader.getOriginatingProvider();
            String[] suffixes = readerSpi.getFileSuffixes();
            for (int i = 0; i < suffixes.length; i++) {
                if (suffix == null || suffix.length() > suffixes[i].length()) {
                    suffix = suffixes[i].toLowerCase(Locale.US);
                }
            }
        }
        return suffix;
    }
    
    /**
     * ԂcontentTypeÃNXœǂݍ݉\ǂ𔻒肵܂B
     * 
     * @param in mFΏۂƂȂXg[
     * @return ǂݍ݉\łꍇtrue
     * @throws IOException ͗Oꍇ
     */
    public static boolean checkReadableContentType(
            InputStream in) throws IOException {
        if (getDefaultFileSuffix(in) != null) {
            return true;
        }
        return false;
    }
    
    /**
     * FileBufferedImageɕϊĕԂ܂B
     * ϊoȂꍇnullԂ܂B
     * 
     * @param file ϊΏۂFile
     * @return fileϊBufferedImage or null
     * @throws IOException o̓G[̏ꍇ
     * @throws IllegalArgumentException ǂݍ߂摜t@CłȂꍇ
     */
    public static BufferedImage convertFileToBufferedImage(
            File file) throws IOException {
        try {
            return convertStreamToBufferedImage(new FileInputStream(file));
        } catch (IllegalArgumentException e) {
            throw new IllegalArgumentException(
                    "File is not a picture file. file : " + file, e);
        }
    }
    
    /**
     * InputStreamBufferedImageɕϊĕԂ܂B
     * ϊoȂꍇnullԂ܂B
     * 
     * @param in ϊΏۂInputStream
     * @return inϊBufferedImage or null
     * @throws IOException o̓G[̏ꍇ
     * @throws IllegalArgumentException 
     *          ǂݍ߂`̉摜Xg[ł͂Ȃꍇ
     */
    public static BufferedImage convertStreamToBufferedImage(
            InputStream in) throws IOException {
        BufferedInputStream bis = null;
        try {
            bis = new BufferedInputStream(in);
            BufferedImage origin = _readImageStream(bis);
            if (origin == null) {
                return null;
            }
            // IWî܂܎g킸A炭H₷ftHǧ`ɕϊB
            // ̑ΉŁA傫ȃAbv[h摜ꍇɌIɃptH[}XPB
            // J[fς̂Ȃ̂A኱FςČB
            BufferedImage result =
                new BufferedImage(origin.getWidth(), origin.getHeight(), BufferedImage.TYPE_INT_RGB);
            Graphics g = result.getGraphics();
            g.drawImage(origin, 0, 0, null);
            g.dispose();
            return result;
        } finally {
            if (bis != null) {
                bis.close();
            }
        }
    }
    
    // CMYK, RGB𔻒fāAO𔭐Ȃ悤JPG̃Xg[BufferedImageɕϊ܂B
    // CMYK̖Fhttp://bugs.sun.com/bugdatabase/view_bug.do;:YfiG?bug_id=5100094
    // ̎QlFhttp://forum.java.sun.com/thread.jspa?threadID=774288&tstart=135
    // ICCvt@CJPGłÃvt@CĂȂׁAϊɂ͗pȂ
    private static BufferedImage _readImageStream(InputStream source) throws IOException {
        ImageInputStream input = null;
        ImageReader reader = null;
        try {
            input = ImageIO.createImageInputStream(source);
            Iterator<ImageReader> iterator = ImageIO.getImageReaders(input);
            if (iterator == null || !iterator.hasNext()) {
                System.err.println("InputStream is not a readable picture stream.");
                return null;
            }
            reader = iterator.next();
            reader.setInput(input);
            String format = reader.getFormatName();
            if ("JPEG".equalsIgnoreCase(format) || "JPG".equalsIgnoreCase(format)) {
                // adobě`ꍇAD悷
				IIOMetadata metadata = null;
				try {
					metadata = reader.getImageMetadata(0);
				} catch (IIOException e) {
					// MetaDataǂݍ߂Ȃ́AJREReaderɔC
					// (getImageTypeeŕԂŏImageTypeSpecifierŃtH[}bg)
					// ݂킩ĂA̗O:
					// 1) }[J APP2(װ̧) ݂̂ APP0(JFIF) ȂꍇB
					// 2) SOI ̒ APP0 ȂAԂɑ̃}[JłꍇB
			        return readImageReader(reader);
				}
                String metadataFormat = metadata.getNativeMetadataFormatName();
                IIOMetadataNode metadataNode = (IIOMetadataNode) metadata.getAsTree(metadataFormat);
                Raster raster = reader.readRaster(0, reader.getDefaultReadParam());
                int numFrameComponents = _getJpgNumFrameComponents(metadataNode, raster);
                NodeList children = metadataNode.getElementsByTagName("app14Adobe");
                if (children.getLength() > 0) {
                    metadataNode = (IIOMetadataNode) children.item(0);
                    int transform = Integer.parseInt(metadataNode.getAttribute("transform"));
                    ColorConverter converter;
                    if (transform == 2) {
                        converter = new YCbCrKColorConverter(raster);
                    } else if (transform == 1) {
                        converter = new YCbCrColorConverter(raster);
                    } else {
                        // transform == 0
                        if (numFrameComponents == 4) {
                            converter = new CMYKColorConverter(raster);
                        } else if (numFrameComponents == 1 || numFrameComponents == 2) {
                        	// O[XP[i2̓At@܂܂Ă邪Ał炵j
                        	converter = new GlayScaleColorConverter(raster);
                        } else {
                            // numFrameComponents == 3
                            converter = new RGBColorConverter(raster);
                        }
                    }
                    return _convertRasterImage(converter);
                } else {
                    ColorConverter converter;
                    if (numFrameComponents == 1 || numFrameComponents == 2) {
                        // O[XP[i2̓At@܂܂Ă邪Ał炵j
                        converter = new GlayScaleColorConverter(raster);
                    } else {
                        // numFrameComponents == 3 || numFrameComponents == 4
                        // YCbCri4̓At@܂܂Ă邪Ał炵j
                        converter = new YCbCrColorConverter(raster);
                    }
                    return _convertRasterImage(converter);
                }
            }
            // ̑̉摜`JREReaderɔC
            return readImageReader(reader);
        } finally {
            if (input != null) {
                try {
                    input.close();
                } catch (IOException ignore) {
                    // ǂ鎖oȂ̂Ŗ
                }
            }
            if (reader != null) {
                reader.dispose();
            }
        }
    }
    
    private static BufferedImage readImageReader(ImageReader reader) throws IOException {
        // IllegalArgumentExceptionȂǂ̃G[ꍇ̂
        // ł܂Ƃ߂catch悤ɂĂB
        // http://schmidt.devlib.org/java/creating-thumbnails-with-java.html
        try {
            return reader.read(0);
        } catch (ArrayIndexOutOfBoundsException e) {
            e.printStackTrace();
            return null;
        } catch (IIOException e) {
            e.printStackTrace();
            return null;
        } catch (CMMException e) {
            e.printStackTrace();
            return null;
        } catch (IllegalArgumentException e) {
            e.printStackTrace();
            return null;
        }
    }
    
    private static int _getJpgNumFrameComponents(IIOMetadataNode metadataNode, Raster raster) {
        // t[wb_[ZOgR|[lg̎擾݂
        NodeList children = metadataNode.getElementsByTagName("sof");
        if (children.getLength() > 0) {
            metadataNode = (IIOMetadataNode) children.item(0);
            String value = metadataNode.getAttribute("numFrameComponents");
            if (value != null && NumberUtil.isInteger(value)) {
                return NumberUtil.parseInt(value);
            }
        }
        // _ȂRaster̃ohԂ
        return raster.getNumBands();
    }
    
    /**
     * FϊsۃNXłB
     * 
     * @author Shinobu Aoki
     */
    private static abstract class ColorConverter {
        protected final Raster target;
        protected final int w;
        protected final int h;
        protected ColorConverter(Raster target) {
            this.target = target;
            this.w = target.getWidth();
            this.h = target.getHeight();
        }
        /**
         * targetRGB`̔zɕϊ܂B
         * 
         * @return ϊς݂RGB`̔z
         */
        protected abstract byte[] convert();
        
        int getWidth() {
            return w;
        }
        int getHeight() {
            return h;
        }
        Raster getRaster() {
            return target;
        }
    }
    
    /**
     * YCbCrK̐FԂŒ`ꂽRasterRGB`̔zɕϊNXłB<br />
     * <a href="http://hp.vector.co.jp/authors/VA032610/JPEGFormat/colorspace.htm" target="_blank">Qly[W</a>
     * 
     * @author Shinobu Aoki
     */
    private static class YCbCrKColorConverter extends ColorConverter {
        private YCbCrKColorConverter(Raster target) {
            super(target);
        }
        protected byte[] convert() {
            byte[] rgb = new byte[w * h * 3];
            float[] Y = target.getSamples(0, 0, w, h, 0, (float[]) null);
            float[] Cb = target.getSamples(0, 0, w, h, 1, (float[]) null);
            float[] Cr = target.getSamples(0, 0, w, h, 2, (float[]) null);
            float[] K = target.getSamples(0, 0, w, h, 3, (float[]) null);
            for (int i = 0, imax = Y.length, base = 0; i < imax; i++, base += 3) {
                float k = 220 - K[i], y = 255 - Y[i], cb = 255 - Cb[i], cr = 255 - Cr[i];
                
                double val = y + 1.402 * (cr - 128) - k;
                val = (val - 128) * 0.65f + 128;
                rgb[base] = val < 0.0 ? (byte) 0 : val > 255.0 ? (byte) 0xff : (byte) (val + 0.5);
                
                val = y - 0.34414 * (cb - 128) - 0.71414 * (cr - 128) - k;
                val = (val - 128) * 0.65f + 128;
                rgb[base + 1] = val < 0.0 ? (byte) 0 : val > 255.0 ? (byte) 0xff : (byte) (val + 0.5);
                
                val = y + 1.772 * (cb - 128) - k;
                val = (val - 128) * 0.65f + 128;
                rgb[base + 2] = val < 0.0 ? (byte) 0 : val > 255.0 ? (byte) 0xff : (byte) (val + 0.5);
            }
            return rgb;
        }
    }
    
    /**
     * YCbCr̐FԂŒ`ꂽRasterRGB`̔zɕϊNXłB<br />
     * <a href="http://hp.vector.co.jp/authors/VA032610/JPEGFormat/marker/APP14Adobe.htm" target="_blank">Qly[W</a>
     * 
     * @author Shinobu Aoki
     */
    private static class YCbCrColorConverter extends ColorConverter {
        private YCbCrColorConverter(Raster target) {
            super(target);
        }
        protected byte[] convert() {
            byte[] rgb = new byte[w * h * 3];
            float[] Y = target.getSamples(0, 0, w, h, 0, (float[]) null);
            float[] Cb = target.getSamples(0, 0, w, h, 1, (float[]) null);
            float[] Cr = target.getSamples(0, 0, w, h, 2, (float[]) null);
            int len = Y.length;
            for (int i = 0, imax = len, base = 0; i < imax; i++, base += 3) {
                float y = Y[i], cb = Cb[i], cr = Cr[i];
                
                double val = y + 1.402 * (cr - 128);
                rgb[base] = val < 0.0 ? (byte) 0 : val > 255.0 ? (byte) 0xff : (byte) (val + 0.5);
                
                val = y - 0.34414 * (cb - 128) - 0.71414 * (cr - 128);
                rgb[base + 1] = val < 0.0 ? (byte) 0 : val > 255.0 ? (byte) 0xff : (byte) (val + 0.5);
                
                val = y + 1.772 * (cb - 128);
                rgb[base + 2] = val < 0.0 ? (byte) 0 : val > 255.0 ? (byte) 0xff : (byte) (val + 0.5);
            }
            return rgb;
        }
    }
    
    /**
     * PxiYj̔摜RasterRGB`̔zɕϊNXłB<br />
     * <a href="http://hp.vector.co.jp/authors/VA032610/JPEGFormat/colorspace.htm" target="_blank">Qly[W</a>
     * 
     * @author Shinobu Aoki
     */
    private static class GlayScaleColorConverter extends ColorConverter {
        private GlayScaleColorConverter(Raster target) {
            super(target);
        }
        protected byte[] convert() {
            byte[] rgb = new byte[w * h * 3];
            int[] Y = target.getSamples(0, 0, w, h, 0, (int[]) null);
            for (int i = 0, len = Y.length, base = 0; i < len; i++, base += 3) {
                byte val = (byte) Y[i];
                rgb[base] = val;
                rgb[base + 1] = val;
                rgb[base + 2] = val;
            }
            return rgb;
        }
    }
    
    private static class CMYKColorConverter extends ColorConverter {
        private CMYKColorConverter(Raster target) {
            super(target);
        }
        protected byte[] convert() {
            byte[] rgb = new byte[w * h * 3];
            int[] C = target.getSamples(0, 0, w, h, 0, (int[]) null);
            int[] M = target.getSamples(0, 0, w, h, 1, (int[]) null);
            int[] Y = target.getSamples(0, 0, w, h, 2, (int[]) null);
            int[] K = target.getSamples(0, 0, w, h, 3, (int[]) null);
            for (int i = 0, imax = C.length, base = 0; i < imax; i++, base += 3) {
                int c = 255 - C[i];
                int m = 255 - M[i];
                int y = 255 - Y[i];
                int k = 255 - K[i];
                float kk = k / 255f;
                
                rgb[base] = (byte) (255 - Math.min(255f, c * kk + k));
                rgb[base + 1] = (byte) (255 - Math.min(255f, m * kk + k));
                rgb[base + 2] = (byte) (255 - Math.min(255f, y * kk + k));
            }
            return rgb;
        }
    }
    
    private static class RGBColorConverter extends ColorConverter {
        private RGBColorConverter(Raster target) {
            super(target);
        }
        protected byte[] convert() {
            byte[] rgb = new byte[w * h * 3];
            int[] R = target.getSamples(0, 0, w, h, 0, (int[]) null);
            int[] G = target.getSamples(0, 0, w, h, 1, (int[]) null);
            int[] B = target.getSamples(0, 0, w, h, 2, (int[]) null);
            int len = R.length;
            for (int i = 0, imax = len, base = 0; i < imax; i++, base += 3) {
                rgb[base] = (byte) R[i];
                rgb[base + 1] = (byte) G[i];
                rgb[base + 2] = (byte) B[i];
            }
            return rgb;
        }
    }
    
    /**
     * e̐FԂRGB`ɕϊColorConverterp
     * RasterIuWFNgBufferedImageւ̕ϊs܂B
     * 
     * @param converter ϊsColorConverterIuWFNg
     * @return ϊsBufferedImageIuWFNg
     */
    private static BufferedImage _convertRasterImage(ColorConverter converter) {
        byte[] rgb = converter.convert();
        Raster raster =
            Raster.createInterleavedRaster(new DataBufferByte(rgb, rgb.length),
                    converter.getWidth(), converter.getHeight(), converter.getWidth() * 3, 3,
                    new int[] { 0, 1, 2 }, null);
        ColorSpace cs = ColorSpace.getInstance(ColorSpace.CS_sRGB);
        ColorModel cm = new ComponentColorModel(cs, false, true, Transparency.OPAQUE, DataBuffer.TYPE_BYTE);
        return new BufferedImage(cm, (WritableRaster) raster, true, null);
    }

    /**
     * originImageheightAwidthő啝Ƃďcۂ܂܃TCYA
     * SupportedImageTypě`targetɏo܂B<br/>
     * 傫height, widthȉ̏ꍇAg債AłȂꍇAk܂B
     * 
     * @param origin IWiBufferedImage
     * @param target oFile
     * @param outputType o͂摜̌`
     * @param width ő̕isNZj
     * @param height ő̍isNZj
     * @throws IOException o͗O
     */
    public static void createResizedImage(BufferedImage origin, File target,
            SupportedImageType outputType, int width, int height) throws IOException {
        createResizedImage(origin, target, outputType, width, height, DEFAULT_JPG_COMP_QUALITY);
    }
    
    /**
     * originImageheightAwidthő啝Ƃďcۂ܂܃TCYA
     * SupportedImageTypě`targetɏo܂B<br/>
     * 傫height, widthȉ̏ꍇAg債AłȂꍇAk܂B
     * 
     * @param origin IWiBufferedImage
     * @param target oFile
     * @param outputType o͂摜̌`
     * @param width ő̕isNZj
     * @param height ő̍isNZj
     * @param quality JPGŏo͂ꍇ̈kii0.0`1.0j
     * @throws IOException o͗O
     */
    public static void createResizedImage(BufferedImage origin, File target,
            SupportedImageType outputType, int width, int height, float quality) throws IOException {
        createResizedImage(origin, new BufferedOutputStream(new FileOutputStream(target)),
            outputType, width, height, quality);
    }
    
    /**
     * originheightAwidthő啝Ƃ
     * cۂ܂܃TCYAtypě`targetɏo܂B<br/>
     * 傫heightAwidthȉ̏ꍇ͊g債AłȂꍇ͏k܂B
     * JPG̈ki0.75ŏo͂܂B
     * 
     * @param origin IWiBufferedImage
     * @param target oOutputStream
     * @param outputType o͂摜̌`
     * @param width ő̕isNZj
     * @param height ő̍isNZj
     * @throws IOException o͗O
     */
    public static void createResizedImage(BufferedImage origin, OutputStream target,
            SupportedImageType outputType, int width, int height) throws IOException {
        createResizedImage(origin, target, outputType, width, height, DEFAULT_JPG_COMP_QUALITY);
    }
    
    /**
     * originheightAwidthő啝Ƃ
     * cۂ܂܃TCYAtypě`targetɏo܂B<br/>
     * 傫heightAwidthȉ̏ꍇ͊g債AłȂꍇ͏k܂B
     * 
     * @param origin IWiBufferedImage
     * @param target oOutputStream
     * @param outputType o͂摜̌`
     * @param width ő̕isNZj
     * @param height ő̍isNZj
     * @param quality JPGŏo͂ꍇ̈kii0.0`1.0j
     * @throws IOException o͗O
     * @throws IllegalArgumentException quality0.01.0傫ꍇ
     */
    public static void createResizedImage(BufferedImage origin, OutputStream target,
            SupportedImageType outputType, int width, int height, float quality) throws IOException {
        Dimension rect =
            resize(new Dimension(origin.getWidth(), origin.getHeight()), new Dimension(width, height));
        _createImage(origin, target, rect, outputType, DEFAULT_JPG_COMP_QUALITY);
    }
    
    /**
     * originImageheightAwidthő啝Ƃ
     * cۂ܂܃TCYAjpgƂtargetɏo܂B<br/>
     * 傫height, widthȉ̏ꍇAg債AłȂꍇAk܂B
     * JPG̈ki0.75ŏo͂܂B
     * 
     * @param origin IWiBufferedImage
     * @param target oFile
     * @param width ő̕isNZj
     * @param height ő̍isNZj
     * @throws IOException o͗O
     */
    public static void createResizedJPGImage(BufferedImage origin,
            File target, int width, int height) throws IOException {
        createResizedJPGImage(origin, target, width, height,
            DEFAULT_JPG_COMP_QUALITY);
    }
    
    /**
     * originImageheightAwidthő啝Ƃ
     * cۂ܂܃TCYAjpgƂtargetɏo܂B<br/>
     * 傫height, widthȉ̏ꍇAg債AłȂꍇAk܂B
     * 
     * @param origin IWiBufferedImage
     * @param target oFile
     * @param width ő̕isNZj
     * @param height ő̍isNZj
     * @param quality JPG̈kii0.0`1.0j
     * @throws IOException o͗O
     * @throws IllegalArgumentException quality̒l0.01.0傫ꍇ
     */
    public static void createResizedJPGImage(BufferedImage origin, File target, int width,
            int height, float quality) throws IOException {
        createResizedJPGImage(origin, new BufferedOutputStream(new FileOutputStream(target)),
            width, height, quality);
    }
    
    /**
     * originheightAwidthő啝Ƃ
     * cۂ܂܃TCYAjpgƂtargetɏo܂B<br/>
     * 傫heightAwidthȉ̏ꍇ͊g債AłȂꍇ͏k܂B
     * JPG̈ki0.75ŏo͂܂B
     * 
     * @param origin IWiBufferedImage
     * @param target oOutputStream
     * @param width ő̕isNZj
     * @param height ő̍isNZj
     * @throws IOException o͗O
     */
    public static void createResizedJPGImage(BufferedImage origin, 
            OutputStream target, int width, int height) throws IOException {
        createResizedJPGImage(origin, target, width, height,
            DEFAULT_JPG_COMP_QUALITY);
    }
    
    /**
     * originheightAwidthő啝Ƃ
     * cۂ܂܃TCYAjpgƂtargetɏo܂B<br/>
     * 傫heightAwidthȉ̏ꍇ͊g債AłȂꍇ͏k܂B
     * 
     * @param origin IWiBufferedImage
     * @param target oOutputStream
     * @param width ő̕isNZj
     * @param height ő̍isNZj
     * @param quality JPG̈kii0.0`1.0j
     * @throws IOException o͗O
     * @throws IllegalArgumentException quality0.01.0傫ꍇ
     */
    public static void createResizedJPGImage(BufferedImage origin, OutputStream target, int width,
            int height, float quality) throws IOException {
        Dimension rect =
            resize(new Dimension(origin.getWidth(), origin.getHeight()), new Dimension(width, height));
        _createJPGImage(origin, target, rect, quality);
    }
    
    /**
     * originheightAwidth̑傫ɃTCYāAtypě`targetɏo܂B<br />
     * {@link #createResizedImage(BufferedImage, File, SupportedImageType, int, int)}Ƃ̈Ⴂ́A
     * ̃\bh͏cۂɋIɎw̑傫ɕύX_łB
     * cۂăTCYꍇ͏L̃\bh𗘗pĉB
     * 
     * @param origin IWiBufferedImage
     * @param target oFile
     * @param type o͂摜̌`
     * @param width ő̕isNZj
     * @param height ő̍isNZj
     * @throws IOException o͗O
     */
    public static void createFixsizeImage(BufferedImage origin, File target,
            SupportedImageType type, int width, int height) throws IOException {
        _createImage(origin, new BufferedOutputStream(new FileOutputStream(target)), new Dimension(
                width, height), type, DEFAULT_JPG_COMP_QUALITY);
    }

    /**
     * originheightAwidth̑傫ɃTCYāAtypě`targetɏo܂B<br />
     * {@link #createResizedImage(BufferedImage, File, SupportedImageType, int, int)}Ƃ̈Ⴂ́A
     * ̃\bh͏cۂɋIɎw̑傫ɕύX_łB
     * cۂăTCYꍇ͏L̃\bh𗘗pĉB
     * 
     * @param origin IWiBufferedImage
     * @param target oFile
     * @param type o͂摜̌`
     * @param width ő̕isNZj
     * @param height ő̍isNZj
     * @param quality JPGŏo͂ꍇ̈kii0.0`1.0j
     * @throws IOException o͗O
     */
    public static void createFixsizeImage(BufferedImage origin, File target,
            SupportedImageType type, int width, int height, float quality) throws IOException {
        _createImage(origin, new BufferedOutputStream(new FileOutputStream(target)), new Dimension(
                width, height), type, quality);
    }
    
    /**
     * originheightAwidth̑傫ɃTCYāAtypě`targetɏo܂B<br />
     * {@link #createResizedImage(BufferedImage, OutputStream, SupportedImageType, int, int)}Ƃ̈Ⴂ́A
     * ̃\bh͏cۂɋIɎw̑傫ɕύX_łB
     * cۂăTCYꍇ͏L̃\bh𗘗pĉB
     * 
     * @param origin IWiBufferedImage
     * @param target oOutputStream
     * @param type o͂摜̌`
     * @param width ő̕isNZj
     * @param height ő̍isNZj
     * @throws IOException o͗O
     */
    public static void createFixsizeImage(BufferedImage origin, OutputStream target,
            SupportedImageType type, int width, int height) throws IOException {
        _createImage(origin, target, new Dimension(width, height), type, DEFAULT_JPG_COMP_QUALITY);
    }

    /**
     * originheightAwidth̑傫ɃTCYāAtypě`targetɏo܂B<br />
     * {@link #createResizedImage(BufferedImage, OutputStream, SupportedImageType, int, int)}Ƃ̈Ⴂ́A
     * ̃\bh͏cۂɋIɎw̑傫ɕύX_łB
     * cۂăTCYꍇ͏L̃\bh𗘗pĉB
     * 
     * @param origin IWiBufferedImage
     * @param target oOutputStream
     * @param type o͂摜̌`
     * @param width ő̕isNZj
     * @param height ő̍isNZj
     * @param quality JPGŏo͂ꍇ̈kii0.0`1.0j
     * @throws IOException o͗O
     */
    public static void createFixsizeImage(BufferedImage origin, OutputStream target,
            SupportedImageType type, int width, int height, float quality) throws IOException {
        _createImage(origin, target, new Dimension(width, height), type, quality);
    }
    
    /**
     * originheightAwidth̑傫ɃTCY
     * jpgƂtargetɏo܂B<br />
     * {@link #createResizedJPGImage(BufferedImage, File, int, int)}Ƃ̈Ⴂ́A
     * ̃\bh͏cۂɋIɎw̑傫ɕύX_łB
     * cۂăTCYꍇ͏L̃\bh𗘗pĉB
     * JPG̈ki0.75ŏo͂܂B
     * 
     * @param origin IWiBufferedImage
     * @param target oFile
     * @param width ő̕isNZj
     * @param height ő̍isNZj
     * @throws IOException o͗O
     */
    public static void createFixsizeJPGImage(BufferedImage origin, File target, int width,
            int height) throws IOException {
        _createJPGImage(origin, new BufferedOutputStream(new FileOutputStream(target)), new Dimension(
                width, height), DEFAULT_JPG_COMP_QUALITY);
    }
    
    /**
     * originheightAwidth̑傫ɃTCY
     * jpgƂtargetɏo܂B<br />
     * {@link #createResizedJPGImage(BufferedImage, File, int, int)}Ƃ̈Ⴂ́A
     * ̃\bh͏cۂɋIɎw̑傫ɕύX_łB
     * cۂăTCYꍇ͏L̃\bh𗘗pĉB
     * 
     * @param origin IWiBufferedImage
     * @param target oFile
     * @param width ő̕isNZj
     * @param height ő̍isNZj
     * @param quality JPG̈kii0.0`1.0j
     * @throws IOException o͗O
     * @throws IllegalArgumentException quality0.01.0傫ꍇ
     */
    public static void createFixsizeJPGImage(BufferedImage origin, File target, int width,
            int height, float quality) throws IOException {
        _createJPGImage(origin, new BufferedOutputStream(new FileOutputStream(target)), new Dimension(
                width, height), quality);
    }
    
    /**
     * originheightAwidth̑傫ɃTCY
     * jpgƂtargetɏo܂B<br />
     * {@link #createResizedJPGImage(BufferedImage, OutputStream, int, int)}Ƃ̈Ⴂ́A
     * ̃\bh͏cۂɋIɎw̑傫ɕύX_łB
     * cۂăTCYꍇ͏L̃\bh𗘗pĉB
     * JPG̈ki0.75ŏo͂܂B
     * 
     * @param origin IWiBufferedImage
     * @param target oOutputStream
     * @param width ő̕isNZj
     * @param height ő̍isNZj
     * @throws IOException o͗O
     */
    public static void createFixsizeJPGImage(BufferedImage origin, OutputStream target, int width,
            int height) throws IOException {
        _createJPGImage(origin, target, new Dimension(width, height),
            DEFAULT_JPG_COMP_QUALITY);
    }
    
    /**
     * originheightAwidth̑傫ɃTCY
     * jpgƂtargetɏo܂B<br />
     * {@link #createResizedJPGImage(BufferedImage, OutputStream, int, int)}Ƃ̈Ⴂ́A
     * ̃\bh͏cۂɋIɎw̑傫ɕύX_łB
     * cۂăTCYꍇ͏L̃\bh𗘗pĉB
     * 
     * @param origin IWiBufferedImage
     * @param target oOutputStream
     * @param width ő̕isNZj
     * @param height ő̍isNZj
     * @param quality JPG̈kii0.0`1.0j
     * @throws IOException o͗O
     * @throws IllegalArgumentException quality0.01.0傫ꍇ
     */
    public static void createFixsizeJPGImage(BufferedImage origin, OutputStream target, int width,
            int height, float quality) throws IOException {
        _createJPGImage(origin, target, new Dimension(width, height), quality);
    }
    
    private static void _createImage(BufferedImage origin, OutputStream target, Dimension rect,
            SupportedImageType outputType, float quality) throws IOException {
        ImageOutputStream output = null;
        try {
            BufferedImage targetImage;
            if (origin.getWidth() == rect.width && origin.getHeight() == rect.height) {
                targetImage = new BufferedImage(rect.width,
                        rect.height, BufferedImage.TYPE_INT_RGB);
                Graphics g = targetImage.createGraphics();
                // 摜̓ߕ𔒂œhԂ
                g.setColor(Color.WHITE);
                g.fillRect(0, 0, rect.width, rect.height);
                g.drawImage(origin, 0, 0, rect.width, rect.height, null);
                g.dispose();
            } else {
                targetImage = _resizeImage(origin, rect);
            }
            if (outputType == SupportedImageType.JPG) {
                ImageWriter writer = _getJPGImageWriter();
                ImageWriteParam writeParam = writer.getDefaultWriteParam();
                writeParam.setCompressionMode(ImageWriteParam.MODE_EXPLICIT);
                writeParam.setCompressionQuality(quality);
                output = new MemoryCacheImageOutputStream(target);
                writer.setOutput(output);
                writer.write(null, new IIOImage(targetImage, null, null), writeParam);
            } else {
                ImageIO.write(targetImage, outputType.toString(), target);
            }
        } finally {
            if (output != null) {
                try {
                    output.close();
                } catch (IOException ignore) {
                    // 
                }
            }
            try {
                target.close();
            } catch (IOException ignore) {
                // 
            }
        }
    }
    
    private static void _createJPGImage(BufferedImage origin, 
            OutputStream target, Dimension rect, float quality) throws IOException {
        _createImage(origin, target, rect, SupportedImageType.JPG, quality);
    }
    
    private static ImageWriter _getJPGImageWriter() {
        Iterator<ImageWriter> itr = ImageIO.getImageWritersBySuffix("jpg");
        if (!itr.hasNext()) {
            throw new IllegalArgumentException("JPEGpImageWriterȂׁAJPEG`ŏo͏o܂B");
        }
        return itr.next();
    }
    
    // iKIɉ摜TCYāÂ̔{łX[YɃTCY\ɂ郁\bhłB
    // Ɋg͈ł낤ƂƁA{傫肷
    // ƂĂrȂĂ܂ׁÃ\bh쐬B
    private static BufferedImage _resizeImage(BufferedImage source, Dimension rect) {
        final double coefficient = 1.25;
        int curWidth = source.getWidth();
        int curHeight = source.getHeight();
        Image sourceImage = source;
        while (curWidth < rect.width || curHeight < rect.height) {
            curWidth = (int) Math.ceil(curWidth * coefficient);
            curHeight = (int) Math.ceil(curHeight * coefficient);
            BufferedImage tmpImage = new BufferedImage(curWidth, curHeight, BufferedImage.TYPE_INT_RGB);
            sourceImage = sourceImage.getScaledInstance(curWidth, curHeight, Image.SCALE_SMOOTH);
            Graphics2D g = tmpImage.createGraphics();
            g.setColor(Color.WHITE);
            g.fillRect(0, 0, curWidth, curHeight);
            g.drawImage(sourceImage, 0, 0, curWidth, curHeight, null);
            g.dispose();
            sourceImage = tmpImage;
        }
        BufferedImage resultImage = new BufferedImage(rect.width, rect.height, BufferedImage.TYPE_INT_RGB);
        sourceImage = sourceImage.getScaledInstance(rect.width, rect.height, Image.SCALE_SMOOTH);
        Graphics2D g = resultImage.createGraphics();
        g.setColor(Color.WHITE);
        g.fillRect(0, 0, rect.width, rect.height);
        g.drawImage(sourceImage, 0, 0, rect.width, rect.height, null);
        g.dispose();
        return resultImage;
    }
    
    /**
     * originImageheightAwidtĥǂ炩ł
     * 傫ꍇɏcďk܂B<br/>
     * k̃C[WJPGƂtargetɏo܂B
     * JPG̈ki0.75ŏo͂܂B
     * 
     * @param origin IWiBufferedImage
     * @param target oFile
     * @param width ő̕isNZj
     * @param height ő̍isNZj
     * @throws IOException o͗O
     */
    public static void createScaleDownJPGImage(BufferedImage origin, 
            File target, int width, int height) throws IOException {
        createScaleDownJPGImage(origin, new BufferedOutputStream(
                new FileOutputStream(target)), width, height);
    }
    
    /**
     * originImageheightAwidtĥǂ炩ł
     * 傫ꍇɏcďk܂B<br/>
     * k̃C[WJPGƂtargetɏo܂B
     * 
     * @param origin IWiBufferedImage
     * @param target oFile
     * @param width ő̕isNZj
     * @param height ő̍isNZj
     * @param quality JPG̈kii0.0`1.0j
     * @throws IOException o͗O
     * @throws IllegalArgumentException quality0.01.0傫ꍇ
     */
    public static void createScaleDownJPGImage(BufferedImage origin, File target, int width,
            int height, float quality) throws IOException {
        createScaleDownJPGImage(origin, new BufferedOutputStream(new FileOutputStream(target)),
                width, height, quality);
    }
    
    /**
     * originImageheightAwidtĥǂ炩ł
     * 傫ꍇɏcďk܂B<br/>
     * k̃C[WJPGƂtargetɏo܂B
     * JPG̈ki0.75ŏo͂܂B
     * 
     * @param origin IWiBufferedImage
     * @param target oOutputStream
     * @param width ő̕isNZj
     * @param height ő̍isNZj
     * @throws IOException o͗O
     */
    public static void createScaleDownJPGImage(BufferedImage origin, 
            OutputStream target, int width, int height) throws IOException {
        createScaleDownJPGImage(origin, target, width, height,
            DEFAULT_JPG_COMP_QUALITY);
    }
    
    /**
     * originImageheightAwidtĥǂ炩ł
     * 傫ꍇɏcďk܂B<br/>
     * k̃C[WJPGƂtargetɏo܂B
     * 
     * @param origin IWiBufferedImage
     * @param target oOutputStream
     * @param width ő̕isNZj
     * @param height ő̍isNZj
     * @param quality JPG̈kii0.0`1.0j
     * @throws IOException o͗O
     * @throws IllegalArgumentException quality0.01.0傫ꍇ
     */
    public static void createScaleDownJPGImage(BufferedImage origin, OutputStream target,
            int width, int height, float quality) throws IOException {
        Dimension rect = scaleDown(new Dimension(origin.getWidth(), origin.getHeight()), 
                new Dimension(width, height));
        _createJPGImage(origin, target, rect, quality);
    }
    
    /**
     * width, heightƂāAsourcȇ傫𒴂Ăꍇ
     * ۂ܂܏kBufferedImageԂ܂B
     * 
     * @param source k錳ƂȂBufferedImage
     * @param width ő̕isNZj
     * @param height ő̍isNZj
     * @return kBufferedImage
     */
    public static BufferedImage createScaleDownImage(BufferedImage source, int width, int height) {
        Dimension origin = new Dimension(source.getWidth(), source.getHeight());
        Dimension target = scaleDown(origin, new Dimension(width, height));
        return _resizeImage(source, target);
    }
    
    /**
     * cۂ܂܁Awidth, height̑傫킹܂B
     * width, heightQӂƂΑ傫܂B
     * width, heightQӂ̂ǂ炩ł傫Ώ܂B
     * 
     * @param source 傫킹錳ƂȂBufferedImage
     * @param width ő̕isNZj
     * @param height ő̍isNZj
     * @return TCYBufferedImage
     */
    public static BufferedImage createResizeImage(BufferedImage source, int width, int height) {
        Dimension origin = new Dimension(source.getWidth(), source.getHeight());
        Dimension target = resize(origin, new Dimension(width, height));
        return _resizeImage(source, target);
    }
    
    /**
     * target̏cEƂāAĂꍇ
     * ۂ܂܏kRectԂ܂B
     * 
     * @param origin ̋`TCY
     * @param target `TCY
     * @return k`TCY
     */
    private static Dimension scaleDown(Dimension origin, Dimension target) {
        return _resize(origin, target, true);
    }
    
    /**
     * target̏cEƂāAc
     * ۂ܂ܑ傫킹܂B
     * targetQӂƂΑ傫܂B
     * targetQӂ̂ǂ炩ł傫Ώ܂B
     * 
     * @param origin ̋`TCY
     * @param target ύX`TCY
     * @return ύX`TCY
     */
    private static Dimension resize(Dimension origin, Dimension target) {
        return _resize(origin, target, false);
    }
    
    private static Dimension _resize(Dimension origin, Dimension target, boolean scaleDown) {
        Dimension result = new Dimension();
        if (scaleDown && target.height > origin.height && target.width > origin.width) {
            result.height = origin.height;
            result.width = origin.width;
        } else {
            double hRatio = (double) origin.height / target.height;
            double wRatio = (double) origin.width / target.width;
            if (hRatio > wRatio) {
                double aspect = 1 / hRatio;
                result.height = target.height;
                result.width = (int) Math.round(origin.width * aspect);
            } else {
                double aspect = 1 / wRatio;
                result.height = (int) Math.round(origin.height * aspect);
                result.width = target.width;
            }
        }
        // 0ȉɂȂȂ悤ɃK[h
        if (result.height < 1) {
            result.height = 1;
        }
        if (result.width < 1) {
            result.width = 1;
        }
        return result;
    }
}
